# -*- coding: utf8 -*-

from django import forms
from django.conf import settings
from django.core.cache import cache
from django.db import models
from django.utils.html import escape
from django.utils.safestring import mark_safe

from ldap_inp import LdapEleve, LdapCompteCri, LdapFiliere
from ldap_inp.utf_tools import to_alphabetic
from ldap_inp.conn import ConnectionFailure, TimeoutException
from portailuser.fields import PortailUserDbPhoneField
from portailuser.models import Ecole

from ldap_inp import LdapEleveNomPrenom
from ldap_inp.ldapobject import L

import urllib
from PIL import Image
from io import BytesIO
import os
import re
from django.core.checks import Warning, register


# A custom error check to ensure settings.INSCRIPT_PHOTOS_SRC is present on the file system
@register()
def check_dir_presence(app_configs, **kwargs):
    errors = []
    try:
        os.listdir(settings.INSCRIPT_PHOTOS_SRC)
    except FileNotFoundError:
        errors.append(
            Warning(
                "settings.INSCRIPT_PHOTOS_SRC is not present on the file system",
                hint="please verify you deployed properly the application",
                obj=None,
                id="portail.W001",
            )
        )
    return errors


TAILLE_PHOTO_MAX = 120
TAILLE_PHOTO_FINALE = 116, 144

RATIO_PHOTO = float(TAILLE_PHOTO_FINALE[0]) / float(TAILLE_PHOTO_FINALE[1])

GENRES = (
    (1, "Masculin"),
    (2, "Féminin"),
    (3, "Autre"),
)


def get_filieres(ecole):
    cache_data = cache.get("filieres_" + ecole.ident)

    if cache_data is None:
        cache_data = []
        # On considère uniquement des filières actives
        for filiere in LdapFiliere.objects.filter(L(ecole=ecole) & ~L(inactif="TRUE")):
            nom = filiere.displayName
            if len(nom) > 20:
                nom = filiere.shortName
            cache_data.append((filiere.dn, nom))

        cache.set("filieres_" + ecole.ident, cache_data, 3600)

    return cache_data


class Inscription(models.Model):
    # Informations obligatoires
    nom = models.CharField(max_length=64, verbose_name="Nom")
    prenom = models.CharField(max_length=64, verbose_name="Prénom")
    date_naissance = models.DateField(verbose_name="Date de naissance")
    genre = models.SmallIntegerField(choices=GENRES, verbose_name="Genre")

    # Informations école/AE/INP-net
    promo = models.IntegerField(verbose_name="Promo")
    ecole = models.ForeignKey(Ecole, editable=False, on_delete=models.CASCADE)
    filiere = models.CharField(max_length=64, verbose_name="Filière")
    inscrit_ae = models.BooleanField(verbose_name="Inscrit à l’AE", default=False)
    inscrit_frappe = models.BooleanField(verbose_name="Inscrit à la frappe", default=False)
    inscrit_passvie_etudiant = models.BooleanField(verbose_name="Cotisant au PassVie Étudiant", default=False)
    fin_inscrit_passvie_etudiant = models.IntegerField(
        verbose_name="Année d'expiration PassVie Étudiant (si inscrit)",
        blank=True,
        null=True,
    )
    # Informations facultatives
    surnom = models.CharField(max_length=64, blank=True, verbose_name="Surnom (opt)")
    email = models.EmailField(blank=True, verbose_name="Email annexe (opt)")

    adresse = models.TextField(blank=True, verbose_name="Adresse personnelle (opt)")
    adr_parents = models.TextField(blank=True, verbose_name="Adresse parents (opt)")

    tel_mobile = PortailUserDbPhoneField(max_length=32, blank=True, verbose_name="Téléphone mobile (opt)")
    tel_perso = PortailUserDbPhoneField(max_length=32, blank=True, verbose_name="Téléphone perso (opt)")

    fiche_imp = models.BooleanField(editable=False, default=False)

    login = models.CharField(max_length=10, unique=True, blank=False, editable=False)
    mdp = models.CharField(max_length=20, blank=False, editable=False)

    date_ajout = models.DateTimeField(auto_now_add=True, editable=False)

    # etat : (0 - juste crée) ; (1 - ajoutée à la liste, non valide) ;
    #        (2 - en accord avec le ldap du cri) ; (3 - plus dans la liste, validée)
    etat = models.SmallIntegerField(editable=False, default=0)
    ident_cri = models.CharField(max_length=32, editable=False, default="")
    mail_cri = models.EmailField(editable=False, default="")

    def charger_attrs_persos(self):
        for donnee in DonneeFormulaire.objects.filter(inscription=self):
            setattr(self, donnee.ident_champ, donnee.valeur)

    def __str__(self):
        return "{} - {} {}".format(self.login, self.prenom, self.nom)

    def aff_filiere(self):
        return LdapFiliere.objects.getDN(self.filiere).shortName

    def save(self):
        assert self.login and self.mdp and self.ecole
        return super(Inscription, self).save()


class AffichageChamp(object):
    def __init__(self, ident, nom, options_filtrage=None, options_affichage=None):
        self.ident = ident
        self.nom = nom
        self.options_filtrage = options_filtrage
        self.options_affichage = options_affichage
        self.dict_options_filtrage = dict(options_filtrage) if options_filtrage else {}
        self.est_affiche = False
        self.opt_filtrage_act = None

        if not self.options_affichage:
            self.options_affichage = self.dict_options_filtrage

    def recup_parametres(self, params):
        self.est_affiche = ("aff_" + self.ident) in params
        opt_filtrage = params.get("filtre_" + self.ident, None)
        if opt_filtrage not in self.dict_options_filtrage.keys():
            opt_filtrage = None
        self.opt_filtrage_act = opt_filtrage

    def config_parametres(self):
        ret = '<label for="">{} : </label>'.format(escape(self.nom))
        ret += '<label><input type="checkbox" name="aff_{}"{} />'.format(
            escape(self.ident), ' checked="checked"' if self.est_affiche else ""
        )
        ret += " Afficher</label>"
        if self.options_filtrage:
            ret += ' | <label>Filtrer :\n<select name="filtre_{}">\n'.format(escape(self.ident))
            for (ident, nom) in self.options_filtrage:
                if ident == self.opt_filtrage_act:
                    select = ' selected="selected"'
                else:
                    select = ""

                ret += '<option value="{}"{}>{}</option>\n'.format(escape(ident) if ident else "", select, escape(nom))
            ret += "</select></label>"

        return mark_safe(ret)

    def filtrer(self, entree):
        if self.opt_filtrage_act is None:
            return True

        try:
            valeur = str(getattr(entree, self.ident))
        except UnicodeEncodeError:
            valeur = str(getattr(entree, self.ident))
        except AttributeError:
            return False

        return valeur == self.opt_filtrage_act

    def afficher(self, entree):
        try:
            valeur = str(getattr(entree, self.ident))
        except UnicodeEncodeError:
            valeur = str(getattr(entree, self.ident))
        except AttributeError:
            return "<Non spécifié>"

        if valeur in self.options_affichage:
            return self.options_affichage[valeur]

        return valeur


class AffichageChampCaseACocher(AffichageChamp):
    def __init__(self, ident, nom):
        AffichageChamp.__init__(
            self,
            ident,
            nom,
            [(None, "-"), ("True", "Oui"), ("False", "Non")],
            {"True": "X", "False": ""},
        )


class ChampFormulaire(models.Model):
    ident = models.SlugField(max_length=32, verbose_name="Identifiant")
    titre = models.CharField(max_length=256, verbose_name="Titre")
    ecole = models.ForeignKey(Ecole, editable=False, on_delete=models.CASCADE)
    inscription_directe = models.BooleanField(default=False)

    def erreurs_champs(self):
        return None

    def save(self):
        assert self.ecole
        super(ChampFormulaire, self).save()

    def delete(self):
        DonneeFormulaire.objects.filter(inscription__ecole=self.ecole, ident_champ=self.ident).delete()
        super(ChampFormulaire, self).delete()

    def affichage_champ(self):
        return AffichageChamp(self.ident, self.titre)

    class Meta:
        abstract = True
        unique_together = (("ident", "ecole"),)


class ChampCaseACocher(ChampFormulaire):
    type = "cc"
    nom_type = "Case à cocher"

    defaut = models.BooleanField(verbose_name="Cochée par défaut", default=True)

    def ajouter_a(self, form, prec):
        init = self.defaut if prec is None else (True if prec == "True" else False)
        form.fields[self.ident] = forms.BooleanField(required=False, label=mark_safe(self.titre), initial=init)

    def affichage_champ(self):
        return AffichageChampCaseACocher(self.ident, self.titre)


class ChampMenuDeroulant(ChampFormulaire):
    type = "menu"
    nom_type = "Menu déroulant"
    choix_oblig = models.BooleanField(verbose_name="Choix obligatoire", default=False)

    choix = models.TextField(verbose_name="Choix", help_text="(1 par ligne, format « clé:valeur »)")
    defaut = models.CharField(max_length=32, blank=True, verbose_name="Par défaut (clé)")

    def tuple_choix(self):
        liste_choix = []

        if not self.choix_oblig:
            liste_choix.append(("", "--------"))

        for ligne in self.choix.split("\n"):
            if not ligne:
                continue
            res = re.match(r"^([a-zA-Z0-9_-]+):(.+)$", ligne)
            if not res or not res.group(2).strip():
                return None

            liste_choix.append((res.group(1), res.group(2).strip()))

        if len(liste_choix) < 2:
            return None

        return tuple(liste_choix)

    def erreurs_champs(self):
        choix = self.tuple_choix()
        if not choix:
            return {"choix": forms.utils.ErrorList(["Format d’entrée invalide."])}

        if self.defaut and self.defaut not in dict(choix):
            return {"defaut": forms.utils.ErrorList(["Valeur par défaut non présente parmi les clés."])}

        return None

    def ajouter_a(self, form, prec):
        init = (self.defaut if self.defaut else None) if prec is None else prec
        form.fields[self.ident] = forms.ChoiceField(
            required=self.choix_oblig,
            label=self.titre,
            initial=init,
            choices=self.tuple_choix(),
        )

    def affichage_champ(self):
        choix = list(self.tuple_choix())
        if self.choix_oblig:
            choix.insert(0, (None, "-"))
        else:
            choix[0] = (None, "-")

        return AffichageChamp(self.ident, self.titre, choix)


class ChampLigneTexte(ChampFormulaire):
    type = "ltxt"
    nom_type = "Ligne de texte"

    choix_oblig = models.BooleanField(verbose_name="Remplissage obligatoire", default=False)
    defaut = models.CharField(max_length=64, blank=True, verbose_name="Valeur par défaut")

    def ajouter_a(self, form, prec):
        init = (self.defaut if self.defaut else None) if prec is None else prec
        form.fields[self.ident] = forms.CharField(required=self.choix_oblig, label=self.titre, initial=init)


class ChampTexte(ChampFormulaire):
    type = "ctxt"
    nom_type = "Champ de texte"

    choix_oblig = models.BooleanField(verbose_name="Remplissage obligatoire", default=False)
    defaut = models.TextField(blank=True, verbose_name="Valeur par défaut")

    def ajouter_a(self, form, prec):
        init = (self.defaut if self.defaut else None) if prec is None else prec
        form.fields[self.ident] = forms.CharField(
            required=self.choix_oblig,
            label=self.titre,
            initial=init,
            widget=forms.Textarea,
        )


class ChampEntier(ChampFormulaire):
    type = "int"
    nom_type = "Nombre entier"

    choix_oblig = models.BooleanField(verbose_name="Remplissage obligatoire", default=False)
    val_min = models.IntegerField(verbose_name="Valeur minimale")
    val_max = models.IntegerField(verbose_name="Valeur maximale")
    defaut = models.IntegerField(blank=True, verbose_name="Valeur par défaut")

    def erreurs_champs(self):
        if self.val_min >= self.val_max:
            return {"val_min": forms.utils.ErrorList(["La valeur minimale est supérieure à la valeur maximale."])}
        if (self.defaut is not None) and (self.defaut < self.val_min or self.defaut > self.val_max):
            return {"defaut": forms.utils.ErrorList(["Valeur par défaut non comprise entre le minimum et le maximum."])}
        return None

    def ajouter_a(self, form, prec):
        init = (self.defaut if self.defaut else "") if prec is None else prec
        form.fields[self.ident] = forms.IntegerField(
            required=self.choix_oblig,
            label=self.titre,
            initial=init,
            min_value=self.val_min,
            max_value=self.val_max,
        )


CHAMPS_FORM = (
    (ChampCaseACocher.type, ChampCaseACocher.nom_type),
    (ChampMenuDeroulant.type, ChampMenuDeroulant.nom_type),
    (ChampLigneTexte.type, ChampLigneTexte.nom_type),
    (ChampTexte.type, ChampTexte.nom_type),
    (ChampEntier.type, ChampEntier.nom_type),
)

MODELES_CHAMP = {
    ChampCaseACocher.type: ChampCaseACocher,
    ChampMenuDeroulant.type: ChampMenuDeroulant,
    ChampLigneTexte.type: ChampLigneTexte,
    ChampTexte.type: ChampTexte,
    ChampEntier.type: ChampEntier,
}


class DonneeFormulaire(models.Model):
    inscription = models.ForeignKey(Inscription, on_delete=models.CASCADE)
    ident_champ = models.SlugField(max_length=32)
    valeur = models.TextField(null=True, blank=True)

    def __str__(self):
        return "{} {} {}".format(self.inscription, self.ident_champ, self.valeur)

    class Meta:
        unique_together = (("inscription", "ident_champ"),)


def liste_photos():
    liste = []

    for fich in os.listdir(settings.INSCRIPT_PHOTOS_SRC):
        if re.search(r"[\uD800-\uDFFF]", fich) is not None:
            try:
                fich_iso = fich.encode("utf8", "surrogateescape").decode("iso-8859-1")
                assert fich != fich_iso
                os.rename(
                    os.path.join(settings.INSCRIPT_PHOTOS_SRC, fich),
                    os.path.join(settings.INSCRIPT_PHOTOS_SRC, fich_iso),
                )
                fich = fich_iso
            except Exception:
                # Let's ignore this file
                continue
        (nom, ext) = os.path.splitext(fich)
        if ext.lower() in (".jpg", ".jpeg", ".png"):
            full_name = os.path.join(settings.INSCRIPT_PHOTOS_SRC, fich)
            liste.append((os.stat(full_name).st_ctime, fich))
    liste.sort(reverse=True)
    return [i[1] for i in liste]


def reduire_taille(taille, max):
    larg, haut = taille
    if larg < max and haut < max:
        return taille

    if larg < haut:  # Portrait
        larg = max * larg / haut
        haut = max
    else:  # Paysage
        haut = max * haut / larg
        larg = max

    return int(larg), int(haut)


def recadrage(taille):
    larg, haut = taille
    larg_alt = int(round(haut * RATIO_PHOTO))
    haut_alt = int(round(larg / RATIO_PHOTO))

    if larg_alt <= larg:
        decoupe = (larg - larg_alt) / 2
        return (decoupe, 0, larg - decoupe, haut)
    else:
        decoupe = (haut - haut_alt) / 2
        return (0, decoupe, larg, haut - decoupe)


def lire_photo(quoted_name):
    donnees = cache.get("inscr_photo_" + quoted_name)
    if donnees is not None:
        return donnees

    try:
        nom = urllib.parse.unquote(quoted_name)
        photo = Image.open(os.path.join(settings.INSCRIPT_PHOTOS_SRC, nom))
    except IOError:
        return None

    donnees = redimensionner_photo(photo)

    cache.set("inscr_photo_" + quoted_name, donnees, 120)

    return donnees


def redimensionner_photo(photo) -> bytes:
    photo = photo.crop(recadrage(photo.size))
    photo = photo.resize(reduire_taille(photo.size, 144), Image.ANTIALIAS)
    photo = photo.convert("RGB")

    donneesIO = BytesIO()
    photo.save(donneesIO, "JPEG", quality=90)

    return donneesIO.getvalue()


def a_photo_enreg(ident_insc):
    chemin = os.path.join(settings.INSCRIPT_PHOTOS_DST, str(ident_insc) + ".jpg")
    return os.access(chemin, os.F_OK)


def lire_photo_enreg(ident_insc):
    try:
        photo = open(os.path.join(settings.INSCRIPT_PHOTOS_DST, str(ident_insc) + ".jpg"), "rb")
        donnees = photo.read()
    finally:
        photo.close()

    return donnees


def image_photo_enreg(ident_insc):
    photo = Image.open(os.path.join(settings.INSCRIPT_PHOTOS_DST, str(ident_insc) + ".jpg"))
    if photo.size != TAILLE_PHOTO_FINALE:
        photo = photo.resize(TAILLE_PHOTO_FINALE, Image.ANTIALIAS)
    return photo


def selectionner_photos(nom_photo_princ, noms_photos_sec):
    if nom_photo_princ:
        donnees_princ = lire_photo(nom_photo_princ)
        if donnees_princ is None:
            return None
    else:
        donnees_princ = None

    donnees_sec = []
    for nom_photo in noms_photos_sec:
        if nom_photo == nom_photo_princ:
            continue

        donnees = lire_photo(nom_photo)
        if donnees is None:
            return None

        donnees_sec.append((nom_photo, donnees))

    return (nom_photo_princ, donnees_princ, donnees_sec)


def enregistrer_photos(donnees, ident_insc):
    nom_photo_princ, donnees_princ, donnees_sec = donnees

    nom_princ = "{}.jpg".format(ident_insc)
    nom_sec = "{}_{{}}.jpg".format(ident_insc)
    chemin_princ = os.path.join(settings.INSCRIPT_PHOTOS_DST, nom_princ)
    chemin_sec = os.path.join(settings.INSCRIPT_PHOTOS_DST, nom_sec)

    a_supprimer = []

    if nom_photo_princ:
        with open(chemin_princ, "bw") as photo:
            photo.write(donnees_princ)
            a_supprimer.append(nom_photo_princ)

    cpt_nom = 1
    for nom_photo, donnees in donnees_sec:
        while os.access(chemin_sec.format(cpt_nom), os.F_OK):
            cpt_nom += 1

        with open(chemin_sec.format(cpt_nom), "bw") as photo:
            photo.write(donnees)

        a_supprimer.append(nom_photo)

    for nom_photo in a_supprimer:
        try:
            os.remove(os.path.join(settings.INSCRIPT_PHOTOS_SRC, nom_photo))
        except OSError:
            pass


class EtatInscr:
    OK = 0
    AVERT = 1
    ERREUR = 2


class LdapCriError(Exception):
    pass


def get_noms_prenoms():
    noms_prenoms = cache.get("inscr_noms_prenoms", set())

    if not noms_prenoms:
        for i in LdapEleveNomPrenom.objects.all():
            noms_prenoms.add((to_alphabetic(i.nom), to_alphabetic(i.prenom)))
        cache.set("inscr_noms_prenoms", noms_prenoms, 3600)

    return noms_prenoms


def invalidate_noms_prenoms():
    cache.delete("inscr_noms_prenoms")


def preparer_inscription(inscription, tout_revalider=False, avec_cri=True):
    from inscriptions.logins import vers_alphabet

    noms_prenoms = get_noms_prenoms()
    nom_prenom_doublon = (to_alphabetic(inscription.nom), to_alphabetic(inscription.prenom)) in noms_prenoms

    if inscription.prenom.upper() == inscription.prenom or inscription.prenom.lower() == inscription.prenom:
        if inscription.prenom.find(" ") < 0:
            inscription.prenom = inscription.prenom.capitalize()
        else:
            return (EtatInscr.ERREUR, "Capitalisation incorrecte du prénom.")

    if inscription.prenom.strip() != inscription.prenom:
        return (EtatInscr.ERREUR, "Espaces au début/à la fin du prénom.")

    if inscription.nom.upper() == inscription.nom or inscription.nom.lower() == inscription.nom:
        if inscription.nom.find(" ") < 0:
            inscription.nom = inscription.nom.capitalize()
        else:
            return (EtatInscr.ERREUR, "Capitalisation incorrecte du nom.")

    if inscription.nom.strip() != inscription.nom:
        return (EtatInscr.ERREUR, "Espaces au début/à la fin du prénom.")

    if getattr(inscription, "ident_cri", None) is not None:
        if bool(LdapEleve.objects.filter(loginTP=inscription.ident_cri)):
            return (EtatInscr.ERREUR, "Login CRI déjà utilisé dans l'annuaire !")

    if bool(LdapEleve.objects.filter(uid=inscription.login)):
        return (EtatInscr.ERREUR, "Login déjà utilisé ! (Ne devrait pas arriver)")

    if avec_cri and (tout_revalider or (not inscription.ident_cri or not inscription.mail_cri)):
        prenom_simp = vers_alphabet(inscription.prenom, True)
        nom_simp = vers_alphabet(inscription.nom, True)

        eleves_cri = None
        essais = 2
        while eleves_cri is None and essais > 0:
            try:
                eleves_cri = LdapCompteCri.objects.filter(
                    givenName=prenom_simp,
                    sn=nom_simp,
                    groups="cn=student, ou=groups, dc=n7, dc=fr",
                )
            except ConnectionFailure:
                pass
            except TimeoutException:
                pass
            essais -= 1

        if eleves_cri is None:
            raise LdapCriError

        if len(eleves_cri) == 1 and eleves_cri[0].uid and eleves_cri[0].mail:
            inscription.ident_cri = eleves_cri[0].uid
            inscription.mail_cri = eleves_cri[0].mail[0]

        elif len(eleves_cri) > 1:
            message = "Plusieurs choix trouvés dans le LDAP du CRI :<br />\n"
            message += "<br />\n".join([i.uid + " — " + (i.premier_mail or "???") for i in eleves_cri])
            message += "<br />\n Entrez manuellement les données correctes."
            return (EtatInscr.ERREUR, message)

        else:
            return (
                EtatInscr.ERREUR,
                "Impossible de trouver les données du CRI correspondantes à “{}” “{}”.".format(nom_simp, prenom_simp),
            )

    if not email_cri_valide(inscription):
        return (EtatInscr.ERREUR, "Adresse email du CRI étrange...")

    if nom_prenom_doublon:
        return (
            EtatInscr.AVERT,
            "<strong>Couple nom/prénom déjà utilisé (doublon ?)</strong>",
        )

    inscription.save()

    return (EtatInscr.OK, None)


DOMAINES_MAILS_CRI = {
    Ecole.n7.ident: "inp-n7.fr",
    Ecole.a7.ident: "inp-ensiacet.fr",
    Ecole.ensat.ident: "inp-ensat.fr",
}


def email_cri_valide(inscription: Inscription) -> bool:
    prefix = ""
    if inscription.filiere != "ou=mast,ou=filieres,o=n7,dc=etu-inpt,dc=fr":
        prefix = "etu."

    if inscription.mail_cri.endswith("@" + prefix + inscription.ecole.nom_propre.lower() + ".fr"):
        return True

    if inscription.mail_cri.endswith("@etu.inp-toulouse.fr"):
        return True

    if inscription.mail_cri.endswith("@etu.inp-n7.fr"):
        return True

    if inscription.mail_cri.endswith("@etu.toulouse-inp.fr"):
        return True

    if inscription.ecole.ident in DOMAINES_MAILS_CRI:
        if inscription.mail_cri.endswith("@" + prefix + DOMAINES_MAILS_CRI[inscription.ecole.ident]):
            return True

    return False


def donnees_cri(rech_nom, rech_prenom):
    from inscriptions.logins import vers_alphabet

    nom = vers_alphabet(rech_nom, True)
    prenom = vers_alphabet(rech_prenom, True)

    if not nom.replace("*", "") and not prenom.replace("*", ""):
        return []

    if not nom:
        nom = "*"

    if not prenom:
        prenom = "*"

    essais = 2
    eleves_cri = None

    while eleves_cri is None and essais > 0:
        try:
            eleves_cri = LdapCompteCri.objects.filter(givenName__matching=prenom, sn__matching=nom)
        except ConnectionFailure:
            pass
        except TimeoutException:
            pass
        essais -= 1

    if eleves_cri is None:
        return None

    annee = re.compile(r"^cn=new(\d{4}),ou=groups,dc=n7,dc=fr$")

    def calcul_annee(groups):
        for group in groups:
            res = annee.match(group)
            if res is not None:
                return res.group(1)
        return "<pas d'année>"

    return [
        {
            "dn": i.dn,
            "ident": i.uid,
            "mail": i.premier_mail or "<pas de mail>",
            "nom": i.sn,
            "prenom": i.givenName,
            "annee": calcul_annee(i.groups),
        }
        for i in eleves_cri
    ]
